#include <dol.h>
#include "global.h"

typedef double matrix[N][N];

//Transpose
void Transpose(matrix A, matrix B) {
  int i, j;
  for(i = 0; i < N; i++)
  {
    for(j = 0; j < N; j++)
    {
      B[j][i] = A[i][j];
    }
  }
}

//Subtract
void Subtract(matrix A, matrix B)	//TBD : triangular matrix subtract will have any optimization?
{
	int i, j;

	for(i = 0; i < N; i++)
	{
		for(j = 0; j < N; j++)
		{
			A[i][j] = A[i][j] - B[i][j];
		}
	}
}

//Multiply
void Multiply(matrix A, matrix B, matrix C)
{
	int i, j, k;
	double sum;
	for(i = 0; i < N; i++)
	{
		for(j = 0; j < N; j++) 
		{
			sum = 0.0;
			for(k = 0; k < N; k++)
				sum += A[i][k] * B[k][j];
			C[i][j] = sum;
    		}
	}
}

//Transpose & Multiply
void TMultiply(matrix A, matrix C)
{
	int i, j, k;
	double sum;
	for(i = 0; i < N; i++)
	{
		for(j = 0; j < N; j++) 
		{
			sum = 0.0;
			for(k = 0; k < N; k++)
			{
				sum += A[i][k] * A[j][k];
			}
			C[i][j] = sum;
    		}
	}
}

//Cholesky Decomposition
void Cholesky(matrix A, matrix L)
{
	int i, j, k;
	double sum;

	for(i = 0; i < N; i++)
	{
		for(j = 0; j < N; j++)
		{
			L[i][j] = 0.0;
		}
	}

	for(j = 0; j < N; j++)
	{
		sum = 0.0;
		for(k = 0; k < j; k++)
		{
			sum += (L[j][k]*L[j][k]);
		}
		L[j][j] = sqrt(A[j][j] - sum);

		for(i = j+1; i < N; i++)
		{
			sum = 0.0;
			for(k = 0; k < j; k++)
			{
				sum += L[i][k]*L[j][k];
			}
			L[i][j] = (A[i][j] - sum)/L[j][j];
		}
	}
}

//Invert
void Invert(matrix x, matrix L)
{
	int i, j, k;
	double sum;

	for(i = 0; i < N; i++)
	{
		for(j = 0; j < N; j++)
		{
			x[i][j] = 0.0;
		}
	}

	for(k = 0; k < N; k++)
	{
        for(i = k; i < N; i++)
        {
			sum = 0.0;
			if(i == k)
			{
				x[i][k] = 1/L[i][i];
			}
			else
			{
				for(j = k; j <= i-1; j++)
				{
					sum += -(L[i][j]*x[j][k]);
				}
				x[i][k] = sum/L[i][i];
			}
		}
	}
}
