#include <systemc>
#include <list>
#ifdef INCLUDE_PROFILER
#include <stdio.h>
#endif
#include "dol_sched_if.h"

#include "generator_wrapper.h"
#include "consumer_wrapper.h"
#include "p_0_0_wrapper.h"
#include "p_1_1_wrapper.h"
#include "p_1_0_wrapper.h"

using namespace std;

#ifdef INCLUDE_PROFILER
#define PROFILER_OUTPUT_FILENAME "profile.txt"
FILE *profiler_output_file;
unsigned int profiler_event_counter;
#endif

class sc_application : public sc_module 
{
    public:
    SC_HAS_PROCESS(sc_application);

    generator_wrapper generator_ins;
    sc_event generator_event;
    consumer_wrapper consumer_ins;
    sc_event consumer_event;
    p_0_0_wrapper p_0_0_ins;
    sc_event p_0_0_event;
    p_1_1_wrapper p_1_1_ins;
    sc_event p_1_1_event;
    p_1_0_wrapper p_1_0_ins;
    sc_event p_1_0_event;

    sc_event sched_event;
    list<sc_event* > eventList;
    list<sc_event* >::iterator iter;


    Fifo FIFO_GEN_0_0_ins;
    Fifo FIFO_0_0_CON_ins;
    Fifo FIFO_GEN_1_0_ins;
    Fifo FIFO_1_0_CON_ins;
    Fifo FIFO_GEN_1_1_ins;
    Fifo FIFO_1_1_CON_ins;
    Fifo FIFO_0_0_1_0_ins;
    Fifo FIFO_1_0_1_1_ins;

    sc_application(sc_module_name name)
    :       sc_module(name),
    generator_ins("generator"),
    consumer_ins("consumer"),
    p_0_0_ins("p_0_0"),
    p_1_1_ins("p_1_1"),
    p_1_0_ins("p_1_0"),
    FIFO_GEN_0_0_ins("FIFO_GEN_0_0", 1280),
    FIFO_0_0_CON_ins("FIFO_0_0_CON", 1280),
    FIFO_GEN_1_0_ins("FIFO_GEN_1_0", 1280),
    FIFO_1_0_CON_ins("FIFO_1_0_CON", 1280),
    FIFO_GEN_1_1_ins("FIFO_GEN_1_1", 1280),
    FIFO_1_1_CON_ins("FIFO_1_1_CON", 1280),
    FIFO_0_0_1_0_ins("FIFO_0_0_1_0", 1280),
    FIFO_1_0_1_1_ins("FIFO_1_0_1_1", 1280)
    {
        generator_ins.OUTPORT_OUT[0][0] = &FIFO_GEN_0_0_ins;
        p_0_0_ins.INPORT_port_generator = &FIFO_GEN_0_0_ins;
        p_0_0_ins.OUTPORT_port_consumer = &FIFO_0_0_CON_ins;
        consumer_ins.INPORT_IN[0][0] = &FIFO_0_0_CON_ins;
        generator_ins.OUTPORT_OUT[1][0] = &FIFO_GEN_1_0_ins;
        p_1_0_ins.INPORT_port_generator = &FIFO_GEN_1_0_ins;
        p_1_0_ins.OUTPORT_port_consumer = &FIFO_1_0_CON_ins;
        consumer_ins.INPORT_IN[1][0] = &FIFO_1_0_CON_ins;
        generator_ins.OUTPORT_OUT[1][1] = &FIFO_GEN_1_1_ins;
        p_1_1_ins.INPORT_port_generator = &FIFO_GEN_1_1_ins;
        p_1_1_ins.OUTPORT_port_consumer = &FIFO_1_1_CON_ins;
        consumer_ins.INPORT_IN[1][1] = &FIFO_1_1_CON_ins;
        p_0_0_ins.OUTPORT_OUTx1x0 = &FIFO_0_0_1_0_ins;
        p_1_0_ins.INPORT_INx0x0 = &FIFO_0_0_1_0_ins;
        p_1_0_ins.OUTPORT_OUTx1x1 = &FIFO_1_0_1_1_ins;
        p_1_1_ins.INPORT_INx1x0 = &FIFO_1_0_1_1_ins;


        SC_THREAD(thread_init);
        SC_THREAD(thread_sched);
        SC_THREAD(thread_generator);
        SC_THREAD(thread_consumer);
        SC_THREAD(thread_p_0_0);
        SC_THREAD(thread_p_1_1);
        SC_THREAD(thread_p_1_0);
    }

    void thread_init()
    {
        generator_ins.initialize();
        consumer_ins.initialize();
        p_0_0_ins.initialize();
        p_1_1_ins.initialize();
        p_1_0_ins.initialize();
    }

    void thread_sched()
    {
        while (1)
        {
            for (iter=eventList.begin(); iter != eventList.end(); ++iter)
            {
                sc_event* e = (*iter);
                e->notify();
            }
            eventList.clear();
            wait(sched_event);
        }
    }

    void thread_generator()
    {
        while (!generator_ins.isDetached())
        {
            #ifdef INCLUDE_PROFILER
            if (profiler_output_file != NULL) fprintf(profiler_output_file, "%u generator started.\n", profiler_event_counter++);
            #endif
            generator_ins.fire();
            #ifdef INCLUDE_PROFILER
            if (profiler_output_file != NULL) fprintf(profiler_output_file, "%u generator stopped.\n", profiler_event_counter++);
            #endif
            eventList.push_back(&generator_event);
            sched_event.notify();
            wait(generator_event);
        }
    }
    void thread_consumer()
    {
        while (!consumer_ins.isDetached())
        {
            #ifdef INCLUDE_PROFILER
            if (profiler_output_file != NULL) fprintf(profiler_output_file, "%u consumer started.\n", profiler_event_counter++);
            #endif
            consumer_ins.fire();
            #ifdef INCLUDE_PROFILER
            if (profiler_output_file != NULL) fprintf(profiler_output_file, "%u consumer stopped.\n", profiler_event_counter++);
            #endif
            eventList.push_back(&consumer_event);
            sched_event.notify();
            wait(consumer_event);
        }
    }
    void thread_p_0_0()
    {
        while (!p_0_0_ins.isDetached())
        {
            #ifdef INCLUDE_PROFILER
            if (profiler_output_file != NULL) fprintf(profiler_output_file, "%u p_0_0 started.\n", profiler_event_counter++);
            #endif
            p_0_0_ins.fire();
            #ifdef INCLUDE_PROFILER
            if (profiler_output_file != NULL) fprintf(profiler_output_file, "%u p_0_0 stopped.\n", profiler_event_counter++);
            #endif
            eventList.push_back(&p_0_0_event);
            sched_event.notify();
            wait(p_0_0_event);
        }
    }
    void thread_p_1_1()
    {
        while (!p_1_1_ins.isDetached())
        {
            #ifdef INCLUDE_PROFILER
            if (profiler_output_file != NULL) fprintf(profiler_output_file, "%u p_1_1 started.\n", profiler_event_counter++);
            #endif
            p_1_1_ins.fire();
            #ifdef INCLUDE_PROFILER
            if (profiler_output_file != NULL) fprintf(profiler_output_file, "%u p_1_1 stopped.\n", profiler_event_counter++);
            #endif
            eventList.push_back(&p_1_1_event);
            sched_event.notify();
            wait(p_1_1_event);
        }
    }
    void thread_p_1_0()
    {
        while (!p_1_0_ins.isDetached())
        {
            #ifdef INCLUDE_PROFILER
            if (profiler_output_file != NULL) fprintf(profiler_output_file, "%u p_1_0 started.\n", profiler_event_counter++);
            #endif
            p_1_0_ins.fire();
            #ifdef INCLUDE_PROFILER
            if (profiler_output_file != NULL) fprintf(profiler_output_file, "%u p_1_0 stopped.\n", profiler_event_counter++);
            #endif
            eventList.push_back(&p_1_0_event);
            sched_event.notify();
            wait(p_1_0_event);
        }
    }
    #ifdef INCLUDE_PROFILER
    void initialize_profiler()
    {
        if ((profiler_output_file = fopen(PROFILER_OUTPUT_FILENAME,"w"))==NULL)
        {
            printf("Unable to open profiler output file. No profiling output is written.\n");
            return;
        }

        fprintf(profiler_output_file, "c FIFO_GEN_0_0 1280 o generator %pO i p_0_0 %pI\n",(generator_ins.OUTPORT_OUT[0][0]),(p_0_0_ins.INPORT_port_generator));
        fprintf(profiler_output_file, "c FIFO_0_0_CON 1280 o p_0_0 %pO i consumer %pI\n",(p_0_0_ins.OUTPORT_port_consumer),(consumer_ins.INPORT_IN[0][0]));
        fprintf(profiler_output_file, "c FIFO_GEN_1_0 1280 o generator %pO i p_1_0 %pI\n",(generator_ins.OUTPORT_OUT[1][0]),(p_1_0_ins.INPORT_port_generator));
        fprintf(profiler_output_file, "c FIFO_1_0_CON 1280 o p_1_0 %pO i consumer %pI\n",(p_1_0_ins.OUTPORT_port_consumer),(consumer_ins.INPORT_IN[1][0]));
        fprintf(profiler_output_file, "c FIFO_GEN_1_1 1280 o generator %pO i p_1_1 %pI\n",(generator_ins.OUTPORT_OUT[1][1]),(p_1_1_ins.INPORT_port_generator));
        fprintf(profiler_output_file, "c FIFO_1_1_CON 1280 o p_1_1 %pO i consumer %pI\n",(p_1_1_ins.OUTPORT_port_consumer),(consumer_ins.INPORT_IN[1][1]));
        fprintf(profiler_output_file, "c FIFO_0_0_1_0 1280 o p_0_0 %pO i p_1_0 %pI\n",(p_0_0_ins.OUTPORT_OUTx1x0),(p_1_0_ins.INPORT_INx0x0));
        fprintf(profiler_output_file, "c FIFO_1_0_1_1 1280 o p_1_0 %pO i p_1_1 %pI\n",(p_1_0_ins.OUTPORT_OUTx1x1),(p_1_1_ins.INPORT_INx1x0));
    }
    #endif

}
;
int sc_main (int argc, char *argv[])
{
    sc_report_handler::set_actions("/IEEE_Std_1666/deprecated", SC_DO_NOTHING);
    sc_report::register_id(5000, "parameter problem" );
    sc_application my_app_mdl("model1_flattened");
    #ifdef INCLUDE_PROFILER
    my_app_mdl.initialize_profiler();
    #endif
    sc_start(-1,SC_NS);
    #ifdef INCLUDE_PROFILER
    if (profiler_output_file != NULL) fclose(profiler_output_file);
    #endif
    return 0;
}
