#!/bin/bash

##########################################################################
# Run script for Cholesky Decomposition Models
##########################################################################

if [ -n "$5" ]; then
  echo "Usage: ./runexample MODEL_NUM NCUTS DIM LENGTH"
  exit
fi

if [ -z "$1" ]; then
  echo "Usage: ./runexample MODEL_NUM NCUTS DIM LENGTH"
  exit
fi

if [ -z "$2" ]; then
  echo "Usage: ./runexample MODEL_NUM NCUTS DIM LENGTH"
  exit
fi

if [ -z "$3" ]; then
  echo "Usage: ./runexample MODEL_NUM NCUTS DIM LENGTH"
  exit
fi

if [ -z "$4" ]; then
  echo "Usage: ./runexample MODEL_NUM NCUTS DIM LENGTH"
  exit
fi


example=$1
Ncut=$2
DIM=$3
LENGTH=$4

CLASSPATH_DELIMITER=":"

if [[ "${CLASSPATH_DELIMITER}" != ":" &&  "${CLASSPATH_DELIMITER}" != ";" ]]
then
  CLASSPATH_DELIMITER=":"
fi

JARS="./../../../bin/dol.jar${CLASSPATH_DELIMITER}./../../../bin/xercesImpl.jar${CLASSPATH_DELIMITER}./../../../bin/jdom.jar"

HEADER="[RUNMODEL] "

echo 
echo ${HEADER} Create directory "model${example}".
echo 
rm -rf model${example}
mkdir model${example}
cd model${example}

cp ../lib/Xmlgen.class .

echo
echo ${HEADER} Generate the XML
echo
java Xmlgen ${example} ${Ncut} ${DIM}

echo
echo ${HEADER} check compliance for "model${example}_flattened.xml".
echo
java -classpath "${JARS}${CLASSPATH_DELIMITER}./.." dol.helper.validator.XMLValidator model${example}.xml

if [ "$?" != "0" ]; then
  exit
fi


echo
echo ${HEADER} Create flattened XML "model${example}_flattened.xml".
echo
java -classpath "${JARS}${CLASSPATH_DELIMITER}./.." dol.helper.flattener.XMLFlattener model${example}.xml Model${example}Generator
javac Model${example}Generator.java
java Model${example}Generator > model${example}_flattened.xml


echo
echo ${HEADER} Generate C source files.
echo
rm -rf src
mkdir src
cd src
cp ../../lib/Codegen.class .
cp ../../lib/matrix.c .
cp ../../lib/generator.c .
cp ../../lib/generator.h .
cp ../../lib/consumer.c .
cp ../../lib/consumer.h .

java Codegen ${Ncut} ${DIM} ${LENGTH}

cd ../

echo
echo ${HEADER} Run dol.
echo
java -classpath "${JARS}${CLASSPATH_DELIMITER}./.." dol.main.Main -P model${example}_flattened.xml -D model${example}.dot -H systemc


echo
echo ${HEADER} Make SystemC application.
echo
cd systemc/src
make
cp sc_application ./../..
cd ./../..

echo
echo ${HEADER} Run SystemC application.
echo
./sc_application

echo
echo ${HEADER} Exited normally.


